/***************************************************************************
 *   Copyright (C) 2010-2012 by Stephane List                              *
 *   slist@lilotux.net                                                     *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

/*
  Example of arp-scan

root@Ubuntu-desktop:~# arp-scan -l
Interface: eth0, datalink type: EN10MB (Ethernet)
Starting arp-scan 1.6 with 256 hosts (http://www.nta-monitor.com/tools/arp-scan/)
192.168.1.1     00:17:33:1e:2b:74       neuf cegetel
192.168.1.44    00:04:30:32:d7:0c       Netgem
192.168.1.62    00:20:00:62:79:66       LEXMARK INTERNATIONAL, INC.
192.168.1.71    00:40:8c:99:f5:72       AXIS COMMUNICATIONS AB
192.168.1.72    00:1b:67:00:9b:50       Ubiquisys Ltd

5 packets received by filter, 0 packets dropped by kernel
Ending arp-scan 1.6: 256 hosts scanned in 2.339 seconds (109.45 hosts/sec).  5 responded


Example of nmap (v6 min.)
/home/slist/Téléchargements/nmap-6.01/nmap -6 --script=targets-ipv6-multicast-echo.nse

Starting Nmap 6.01 ( http://nmap.org ) at 2013-01-08 10:25 CET
Pre-scan script results:
| targets-ipv6-multicast-echo:
|   IP: fe80::223:aeff:fe71:b779   MAC: 00:23:ae:71:b7:79  IFACE: eth0
|   IP: fe80::210:18ff:fe4e:5d2e   MAC: 00:10:18:4e:5d:2e  IFACE: eth1
|   IP: fe80::202:b3ff:fec0:fb3    MAC: 00:02:b3:c0:0f:b3  IFACE: eth0
|   IP: fe80::280:c8ff:feb9:b745   MAC: 00:80:c8:b9:b7:45  IFACE: eth1
|   IP: fe80::21e:4fff:febf:ac19   MAC: 00:1e:4f:bf:ac:19  IFACE: virbr0
|   IP: fe80::21b:21ff:febe:3ab8   MAC: 00:1b:21:be:3a:b8  IFACE: virbr1
|   IP: fe80::1a03:73ff:febc:df81  MAC: 18:03:73:bc:df:81  IFACE: eth0


*/

#include "dialog.h"
#include "ui_dialog.h"
#include <QStringList>
#include <QProcess>
#include <QFile>
#include <QList>
#include <QByteArray>
#include <QNetworkInterface>
#include <QSettings>

#include <QDebug>

#include "version.h"
#include "test_port.h"
#include "button_http.h"
#include "button_rtsp.h"
#include "button_ssh.h"

//#define DEBUGSLI 1

const quint16 PORT_HTTP = 80;
const quint16 PORT_SSH = 22;
const quint16 PORT_RTSP = 554;

const int col_ipv4 = 0;
const int col_ipv6 = 1;
const int col_ipv6_ll = 2;
const int col_mac = 3;
const int col_vendor = 4;
const int col_http = 5;
const int col_ssh = 6;
const int col_rtsp = 7;
const int col_description = 8;


Dialog::Dialog(QWidget *parent) :
        QDialog(parent),
        ScanProcess(0),
        ui(new Ui::Dialog)

{


    QCoreApplication::setOrganizationName("Arkoon");
    QCoreApplication::setOrganizationDomain("arkoon.net");
    QCoreApplication::setApplicationName("qt-ip-scan");
    QSettings settings;



    ui->setupUi(this);
    setWindowTitle(windowTitle() + " (V" + VERSION + ")");
    reset_button();
    /*
    if (!check_if_ipv4()) {
        ui->ipv4->setEnabled(false);
    }
    if (!check_if_ipv6()) {
        ui->ipv6->setEnabled(false);
    }
    */
}

void Dialog::reset_button()
{
    if (!check_root())
    {
        ui->ScanButton->setText(tr("arp-scan needs root privilege"));
        ui->nmapButton->setText(tr("nmap needs root privilege"));
        ui->ScanButton->setEnabled(false);
        ui->nmapButton->setEnabled(false);
        return;
    }

    if (!check_if_ipv4()) {
        ui->ScanButton->setText(tr("no IPv4 network interface"));
        ui->ScanButton->setEnabled(false);
    } else if (!check_arp_scan()) {
        ui->ScanButton->setText(tr("arp-scan not found"));
        ui->ScanButton->setEnabled(false);
    } else {
        ui->ScanButton->setText(tr("arp-scan"));
        ui->ScanButton->setEnabled(true);
    }

    if (!check_if_ipv6()){
        ui->nmapButton->setText(tr("no IPv6 network interface"));
        ui->nmapButton->setEnabled(true);
    } else if (!check_nmap())
    {
        ui->nmapButton->setText(tr("nmap not found"));
        ui->nmapButton->setEnabled(false);
    } else {
        ui->nmapButton->setText(tr("nmap"));
        ui->nmapButton->setEnabled(true);
    }
}

bool Dialog::check_if_ipv4()
{

    QList<QNetworkInterface> lni = QNetworkInterface::allInterfaces ();
    QNetworkInterface ni;
    bool ipv4_found = false;
    ui->comboBox_ipv4->clear();

    Q_FOREACH(ni, lni) {
        quint8 flag = ni.flags();
        if (flag | QNetworkInterface::IsUp) {
            if ((flag & QNetworkInterface::IsLoopBack) == 0) {
                QNetworkAddressEntry nae;
                bool ipv4 = false;
                Q_FOREACH(nae, ni.addressEntries()) {
                    if (nae.ip().protocol() == QAbstractSocket::IPv4Protocol) {
                        QString color;
                        if (nae.prefixLength()<24) {
                            color = "#ff0000"; // red
                        }
                        else {
                            color = "#000000"; // black
                        }
                        ui->ScanResult->appendHtml("<p>" + ni.name() + " " + nae.ip().toString() + "<span style=\" color:" + color + ";\">"  + QString("/%1").arg(nae.prefixLength()) + "</span></p>");
                        ipv4 = true;
                        ipv4_found = true;
                    }
                }
                if (ipv4) {
                    //qDebug() << nae.ip();
                    ui->comboBox_ipv4->addItem(ni.name());
                }
            }
        }
    }
    // Use eth0 as default interface
    int ethzero = ui->comboBox_ipv4->findText("eth0");
    if (ethzero>=0) {
        ui->comboBox_ipv4->setCurrentIndex(ethzero);
    }
    return ipv4_found;
}

bool Dialog::check_if_ipv6()
{

    QList<QNetworkInterface> lni = QNetworkInterface::allInterfaces ();
    QNetworkInterface ni;
    bool ipv6_found = false;
    ui->comboBox_ipv6->clear();

    Q_FOREACH(ni, lni) {
        quint8 flag = ni.flags();
        if (flag | QNetworkInterface::IsUp) {
            if ((flag & QNetworkInterface::IsLoopBack) == 0) {
                QNetworkAddressEntry nae;
                bool ipv6 = false;
                Q_FOREACH(nae, ni.addressEntries()) {
                    if (nae.ip().protocol() == QAbstractSocket::IPv6Protocol) {
                        QString color;
                        if (nae.ip().toString().startsWith("FE80")) {
                            color = "#000000"; // black
                        } else {
                            color = "#00AA00"; // green
                        }

                        ui->Scan6Result->appendHtml("<span style=\" color:" + color + ";\"><p>" + ni.name() + " " + nae.ip().toString() + QString("/%1").arg(nae.prefixLength()) + "</span></p>");
                        ipv6 = true;
                        ipv6_found = true;
                    }
                }
                if (ipv6) {
                    ui->comboBox_ipv6->addItem(ni.name());
                }
            }
        }
    }
    // Use eth0 as default interface
    int ethzero = ui->comboBox_ipv6->findText("eth0");
    if (ethzero>=0) {
        ui->comboBox_ipv6->setCurrentIndex(ethzero);
    }
    return ipv6_found;
}

bool Dialog::check_root()
{
    QStringList environment = QProcess::systemEnvironment();

    if (environment.contains("USER=root"))
    {
        return true;
    }
    return false;
}

bool Dialog::check_arp_scan()
{
    QFile file("/usr/bin/arp-scan");
    if (file.exists())
        return true;
    else
        return false;
}

bool Dialog::Dialog::check_nmap()
{
    QFile file("/usr/local/bin/nmap");
    if (file.exists())
        return true;
    else
        return false;
}

Dialog::~Dialog()
{
    delete ui;
}

void Dialog::changeEvent(QEvent *e)
{
    QDialog::changeEvent(e);
    switch (e->type()) {
    case QEvent::LanguageChange:
        ui->retranslateUi(this);
        break;
    default:
        break;
    }
}

void Dialog::on_ScanButton_clicked()
{
    if (ScanProcess != NULL)
        return;

    ui->ScanButton->setText(tr("Scanning..."));
    ui->ScanButton->setEnabled(false);
    ui->nmapButton->setEnabled(false);

    /* while (ui->ScanTable->rowCount()>0)
        ui->ScanTable->removeRow(0);
*/

    ScanProcess = new QProcess();
    connect(ScanProcess, SIGNAL(readyReadStandardOutput()), this, SLOT(On_Scan_Output()));
    connect(ScanProcess, SIGNAL(readyReadStandardError()), this, SLOT(On_Scan_Error()));
    connect(ScanProcess, SIGNAL(finished(int)), this, SLOT(On_Scan_Finished(int)));

    QString ni = ui->comboBox_ipv4->currentText();
    ui->ScanResult->appendPlainText("/usr/bin/arp-scan -l -I " + ni);

    ScanProcess->start("/usr/bin/arp-scan", QStringList() << "-l" << "-I" << ni);
}

void Dialog::On_Scan_Output()
{
    QList<QByteArray> list=ScanProcess->readAllStandardOutput().split('\n');
    QByteArray line;

    // for each output line of type :
    // 192.168.1.71    11:12:13:14:15:16       Vendor XYZ
    Q_FOREACH(line, list)
    {
        ui->ScanResult->appendPlainText(line);
        QList<QByteArray> list2=line.split('\t');
        if (list2.size()==3)
        {
            // new table line entry
            int i=ui->ScanTable->rowCount();
            ui->ScanTable->insertRow(i);
            ui->ScanTable->setItem(i, col_ipv4, new QTableWidgetItem(QString(list2[0])));
            ui->ScanTable->setItem(i, col_mac, new QTableWidgetItem(QString(list2[1])));
            ui->ScanTable->setItem(i, col_vendor, new QTableWidgetItem(QString(list2[2])));

            test_port *t1, *t2, *t3;
            t1 = new test_port(this);
            t2 = new test_port(this);
            t3 = new test_port(this);
            connect(t1, SIGNAL(result(const QString &, quint16, bool)), this, SLOT(On_Test_Result(const QString &, quint16, bool)));
            connect(t2, SIGNAL(result(const QString &, quint16, bool)), this, SLOT(On_Test_Result(const QString &, quint16, bool)));
            connect(t3, SIGNAL(result(const QString &, quint16, bool)), this, SLOT(On_Test_Result(const QString &, quint16, bool)));
            t1->run_test(QString(list2[0]),PORT_HTTP);
            t2->run_test(QString(list2[0]),PORT_SSH);
            t3->run_test(QString(list2[0]),PORT_RTSP);

        }
    }
    add_description();

}

void Dialog::On_Scan_Error()
{
    ui->ScanResult->appendPlainText(ScanProcess->readAllStandardError());
    ScanProcess = NULL;
    ui->ScanButton->setText(tr("Scan"));
    ui->ScanButton->setEnabled(true);
    ui->nmapButton->setEnabled(true);


}

void Dialog::On_Scan_Finished(int code)
{
    if (code>=0) {
        ui->ScanResult->appendPlainText(tr("Finished."));
    } else {
        ui->ScanResult->appendPlainText(tr("Finished with errors."));
    }
    ScanProcess = NULL;
    reset_button();
}

void Dialog::On_Test_Result(const QString & ip, quint16 port, bool res)
{
    // search line in table
    for (int i=0;i<ui->ScanTable->rowCount();i++)
    {
        if (res && (ip == ui->ScanTable->item(i, 0)->text()))
        {
            switch (port) {
            case PORT_HTTP:
                {
                    button_http *b = new button_http(this);
                    b->set_ip(ip);
                    ui->ScanTable->setCellWidget(i, col_http, b);
                }
                break;
            case PORT_SSH:
                {
                    button_ssh *b = new button_ssh(this);
                    b->set_ip(ip);
                    ui->ScanTable->setCellWidget(i, col_ssh, b);
                }
                break;
            case PORT_RTSP:
                {
                    button_rtsp *b = new button_rtsp(this);
                    b->set_ip(ip);
                    ui->ScanTable->setCellWidget(i, col_rtsp, b);
                }
                break;
            default:
                qWarning("This port should not be scanned !");
            }
            break;
        }
    }
}
/*
void Dialog::on_ndisc6Button_clicked()
{
    ui->ndisc6Button->setText(tr("Scanning..."));
    ui->ndisc6Button->setEnabled(false);

}
*/
/*
void Dialog::On_Scan6_Output()
{
    QList<QByteArray> list=ScanProcess->readAllStandardOutput().split('\n');
    QByteArray line;

    // for each output line of type :
    // 192.168.1.71    11:12:13:14:15:16       Vendor XYZ
    Q_FOREACH(line, list)
    {
        ui->Scan6Result->appendPlainText(line);
        QList<QByteArray> list2=line.split('\t');
        if (list2.size()==3)
        {
            // new table line entry
            int i=ui->ScanTable->rowCount();
            ui->ScanTable->insertRow(i);
            ui->ScanTable->setItem(i, col_ipv4, new QTableWidgetItem(QString(list2[0])));
            ui->ScanTable->setItem(i, col_mac, new QTableWidgetItem(QString(list2[1])));
            ui->ScanTable->setItem(i, col_vendor, new QTableWidgetItem(QString(list2[2])));

            test_port *t1, *t2, *t3;
            t1 = new test_port(this);
            t2 = new test_port(this);
            t3 = new test_port(this);
            connect(t1, SIGNAL(result(const QString &, quint16, bool)), this, SLOT(On_Test_Result(const QString &, quint16, bool)));
            connect(t2, SIGNAL(result(const QString &, quint16, bool)), this, SLOT(On_Test_Result(const QString &, quint16, bool)));
            connect(t3, SIGNAL(result(const QString &, quint16, bool)), this, SLOT(On_Test_Result(const QString &, quint16, bool)));
            t1->run_test(QString(list2[0]),PORT_HTTP);
            t2->run_test(QString(list2[0]),PORT_SSH);
            t3->run_test(QString(list2[0]),PORT_RTSP);
        }
    }
    for (int j=0; j<3;j++)
    {
        ui->ScanTable->resizeColumnToContents(j);
    }
}

void Dialog::On_Scan6_Error()
{
    ui->Scan6Result->appendPlainText(ScanProcess->readAllStandardError());
}

void Dialog::On_Scan6_Finished(int code)
{
    if (code>=0) {
        ui->Scan6Result->appendPlainText(tr("Finished."));
    } else {
        ui->Scan6Result->appendPlainText(tr("Finished with errors."));
    }
    ui->ndisc6Button->setText(tr("ndisc6"));
    ui->ndisc6Button->setEnabled(true);
}
*/
void Dialog::on_nmapButton_clicked()
{
    if (ScanProcess != NULL) {
        return;
    }

    ui->nmapButton->setText(tr("Scanning..."));
    ui->nmapButton->setEnabled(false);



    QProcessEnvironment env = QProcessEnvironment::systemEnvironment();
    env.insert("LC_ALL", "C"); // Lang=C
    ScanProcess = new QProcess();
    ScanProcess->setProcessEnvironment(env);
    connect(ScanProcess, SIGNAL(readyReadStandardOutput()), this, SLOT(On_nmap_Output()));
    connect(ScanProcess, SIGNAL(readyReadStandardError()), this, SLOT(On_nmap_Error()));
    connect(ScanProcess, SIGNAL(finished(int)), this, SLOT(On_nmap_Finished(int)));


    QString ni = ui->comboBox_ipv6->currentText();
    ui->Scan6Result->appendPlainText("/usr/local/bin/nmap -6 --script=targets-ipv6-multicast-echo.nse -e " + ni);

#ifndef DEBUGSLI

    ScanProcess->start("/usr/local/bin/nmap", QStringList() << "-6" << "--script=targets-ipv6-multicast-echo.nse" << "-e" << ni);
#else

    ScanProcess->start("ls", QStringList() << ".");
#endif
    // Don't use interface
    // ./nmap -6 --script=targets-ipv6-multicast-echo.nse
    /*
/home/slist/Téléchargements/nmap-6.01/nmap -6 --script=targets-ipv6-multicast-echo.nse
/home/slist/Téléchargements/nmap-6.01/nmap -6 --script=targets-ipv6-multicast-echo.nse

Starting Nmap 6.01 ( http://nmap.org ) at 2013-01-08 10:25 CET
Pre-scan script results:
| targets-ipv6-multicast-echo:

|   IP: fe80::223:aeff:fe71:b779   MAC: 00:23:ae:71:b7:79  IFACE: eth0
|   IP: fe80::210:18ff:fe4e:5d2e   MAC: 00:10:18:4e:5d:2e  IFACE: eth1

*/

}


void Dialog::On_nmap_Error()
{
    ui->Scan6Result->appendPlainText(ScanProcess->readAllStandardError());
}

void Dialog::On_nmap_Finished(int code)
{
    if (code==0) {
        ui->Scan6Result->appendPlainText(tr("Finished."));
    } else {
        ui->Scan6Result->appendPlainText(tr("Finished with errors."));
    }
    ScanProcess = NULL;
    reset_button();
}



void Dialog::On_nmap_Output()
{
    QByteArray line;

#ifndef DEBUGSLI
    QList<QByteArray> list=ScanProcess->readAllStandardOutput().split('\n');

    /*
    | targets-ipv6-multicast-echo:

    |   IP: fe80::223:aeff:fe71:b779   MAC: 00:23:ae:71:b7:79  IFACE: eth0
    |   IP: fe80::210:18ff:fe4e:5d2e   MAC: 00:10:18:4e:5d:2e  IFACE: eth1
    */
#else
    QByteArray debug;
    debug.append("| targets-ipv6-multicast-echo:\n");
    debug.append("|   IP: fe80::223:aeff:fe71:b779   MAC: 00:23:ae:71:b7:79  IFACE: eth0\n");
    QList<QByteArray> list=debug.split('\n');
    ui->Scan6Result->appendPlainText("\n\nOutput:\n");
#endif

    Q_FOREACH(line, list)
    {
        QString s(line);
        QString ip;
        QString mac;
        int i;

        if (s.startsWith("|   IP:")) {

            qDebug() << s;

            s.remove(0, sizeof("|   IP:"));
            i=s.indexOf("MAC:");
            ip = s.left(i).trimmed();

            s=s.remove(0,i+sizeof("MAC:"));
            i=s.indexOf("IFACE");
            mac = s.left(i).trimmed();

            ui->Scan6Result->appendPlainText(ip);
            ui->Scan6Result->appendPlainText(mac);

            add_ipv6_entry(ip, mac);
        }

    }
    add_description();
}

int Dialog::find_mac_in_table(const QString & mac)
{
    for (int i=0;i < ui->ScanTable->rowCount();i++) {
        if (ui->ScanTable->item(i,col_mac)->text() == mac)
            return i;
    }
    return -1;
}

void Dialog::add_ipv6_entry(const QString & ipv6,const QString & mac)
{
    int row = find_mac_in_table(mac);
    bool link_local = ipv6.startsWith("fe80:");
    int col_ip;

    if (link_local)
        col_ip = col_ipv6_ll;
    else
        col_ip = col_ipv6;

    if (row==-1) {
        ui->ScanTable->insertRow(0);
        row=0;
        ui->ScanTable->setItem(row, col_mac, new QTableWidgetItem(mac));
    }
    ui->ScanTable->setItem(row, col_ip, new QTableWidgetItem(ipv6));
    ui->ScanTable->resizeColumnToContents(col_mac);
    ui->ScanTable->resizeColumnToContents(col_ip);
}


void Dialog::on_clearTableButton_clicked()
{
    while (ui->ScanTable->rowCount()>0) {
        ui->ScanTable->removeRow(0);
    }
}

void Dialog::on_pushButton_about_clicked()
{
    qApp->aboutQt();
}

void Dialog::closeEvent(QCloseEvent * /*event*/ )
{
    QSettings settings;
    int row;
    qDebug("Closing");

    for (row=0; row<ui->ScanTable->rowCount(); row++)
    {
        settings.beginGroup(ui->ScanTable->item(row, col_mac)->text());
        if (ui->ScanTable->item(row, col_description)) {
            settings.setValue("description", ui->ScanTable->item(row, col_description)->text());
        }
        settings.endGroup();
    }
    settings.sync();
}

void Dialog::add_description()
{
    QSettings settings;
    int row;

    ui->ScanTable->setSortingEnabled(false);
    for (row=0; row<ui->ScanTable->rowCount(); row++)
    {
        foreach (QString mac, settings.childGroups()) {
            if (mac == ui->ScanTable->item(row, col_mac)->text()) {
                settings.beginGroup(mac);
                ui->ScanTable->setItem(row, col_description, new QTableWidgetItem(settings.value("description", "").toString()));
                settings.endGroup();
            }
        }
    }
    ui->ScanTable->setSortingEnabled(true);
    for (int i=0; i<8; i++) {
        ui->ScanTable->resizeColumnToContents(i);
    }
}
